package tetris.d2d;

import java.awt.*;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;

import javax.swing.*;

import tetris.game.*;

public class Display extends JPanel implements KeyListener, GameListener {
	private static final long serialVersionUID = 1L;
	public static final String TITLE = "Tetris 2D";

	private int framesPerSec = 60;
	private int currentFrame = 0;
	private int blockWidth=32, blockHeight=32;

	private GameModel gm;

	public Display(int w, int h) {
		Thread animationThread = new Thread() {
			public void run() {
				while (true) {
					update();
					repaint();
					try {
						Thread.sleep(1000 / framesPerSec);
					} catch (InterruptedException ex) {
					}
				}
			}
		};
		animationThread.start();
		setPreferredSize(new Dimension(blockWidth*w, blockHeight*h));

		gm = new GameModel(this, w, h);
		GameBlock b = gm.randomBlock();
		gm.setActiveBlock(b);
	}

	public void paintComponent(Graphics g) {
		super.paintComponent(g);
		setBackground(Color.WHITE);
		Graphics2D g2d = (Graphics2D) g;

		Color[][] field = gm.getField();
		for (int i = 0; i < field.length; ++i) {
			for (int j = 0; j < field[i].length; ++j) {
				g2d.setColor(field[i][j]);
				g2d.fillRect(j * blockWidth, i * blockHeight, blockWidth, blockHeight);
			}
		}

		GameBlock b = gm.getActiveBlock();
		g2d.setColor(b.color);

		int ax = gm.getActiveX();
		int ay = gm.getActiveY();
		for (int i = 0; i < b.shape.length; ++i) {
			for (int j = 0; j < b.shape[i].length; ++j) {
				if (b.shape[i][j]) {
					g2d.fillRect((ax + j) * blockWidth,
							// Add to y coordinate a fraction of the block height depending on how far into the current second we are
							(int)((ay + i) * blockHeight + blockHeight * (currentFrame / (double)framesPerSec)),
							blockWidth, blockHeight);
				}
			}
		}
	}

	public void update() {
		++currentFrame;
		if (currentFrame >= framesPerSec) {
			gm.moveActiveDown();
			currentFrame = 0;
		}
	}

	public void blockUsedEvent() {
		gm.setActiveBlock(gm.randomBlock());
	}

	public void removeRowEvent(int row) {
	}

	public void gameOverEvent() {
		System.exit(0);
	}

	public void keyPressed(KeyEvent e) {
		if (e.getKeyCode() == KeyEvent.VK_UP
				|| e.getKeyCode() == KeyEvent.VK_KP_UP
				|| e.getKeyCode() == KeyEvent.VK_W) {
			gm.rotateActiveLeft();
		} else if (e.getKeyCode() == KeyEvent.VK_DOWN
				|| e.getKeyCode() == KeyEvent.VK_KP_DOWN
				|| e.getKeyCode() == KeyEvent.VK_S) {
			gm.moveActiveDown();
			currentFrame = 0;
		} else if (e.getKeyCode() == KeyEvent.VK_LEFT
				|| e.getKeyCode() == KeyEvent.VK_KP_LEFT
				|| e.getKeyCode() == KeyEvent.VK_A) {
			gm.moveActiveLeft();
		} else if (e.getKeyCode() == KeyEvent.VK_RIGHT
				|| e.getKeyCode() == KeyEvent.VK_KP_RIGHT
				|| e.getKeyCode() == KeyEvent.VK_D) {
			gm.moveActiveRight();
		}
	}

	public void keyReleased(KeyEvent e) {
	}

	public void keyTyped(KeyEvent e) {
		if (e.getKeyChar() == 'w' || e.getKeyChar() == 'W') {
			gm.rotateActiveLeft();
		} else if (e.getKeyChar() == 's' || e.getKeyChar() == 'S') {
			gm.moveActiveDown();
			currentFrame = 0;
		} else if (e.getKeyChar() == 'a' || e.getKeyChar() == 'A') {
			gm.moveActiveLeft();
		} else if (e.getKeyChar() == 'd' || e.getKeyChar() == 'D') {
			gm.moveActiveRight();
		} else if (e.getKeyChar() == ' ') {
			gm.settleActive();
			currentFrame = 0;
		}
	}

	public static void main(String[] args) {
		SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				JFrame frame = new JFrame(TITLE);
				Display d = new Display(10, 20);
				frame.setContentPane(d);
				frame.pack();
				frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
				frame.setLocationRelativeTo(null);
				frame.setVisible(true);
				frame.addKeyListener(d);
			}
		});
	}
}
