package tetris.d3d;

import javax.media.opengl.*;
import javax.media.opengl.awt.*;
import javax.media.opengl.glu.GLU;
import javax.swing.JFrame;

import java.awt.*;
import java.awt.event.*;
import java.util.ArrayList;

import com.jogamp.opengl.util.*;
import com.jogamp.opengl.util.gl2.GLUT;

import tetris.game.*;

public class Display extends JFrame implements GLEventListener, KeyListener, WindowListener, GameListener {

	private static final long serialVersionUID = 1L;
	
	private int currentFrame = 0;
	
	private GLProfile profile;
	private GLCapabilities capabilities;
	private GLCanvas canvas;
	private GLU glu;
	private GLUT glut;
	private FPSAnimator animator;
	private int framesPerSecond = 60;
	
	//Game variables
	private int width = 20;//number of blocks
	private int height = 20;//-do-
	private double radiusOuter = 8;//outer radius of the playing field
	private double radiusInner = 6;//inner -do-
	private GameModel gm;
	
	//Models
	private ArrayList<Model> models = new ArrayList<Model>();
	//private Model grid;//wireframe grid
	private Model block;//a single block model
	
	//Camera variables
	private double theta = 1.1 * Math.PI;//radians
	private double camRotSpeed = 0.00;
	private double camDistFromOrigo = 75;
	private double camHeight = 45;
	private double[] camLookAt = {0, 0, 25};//what point should the camera look at?
	
	private long score = 0, bonus = 0;

	public Display(int sizeX, int sizeY) {
//		models.add(new Model(true));//test
		gm = new GameModel(this, width, height);
		GameBlock activeBlock = gm.randomBlock();
		gm.setActiveBlock(activeBlock);
		
		//grid = buildGrid();
		block = buildBlock();
		
		setSize(sizeX, sizeY);
		profile = GLProfile.getDefault();
		capabilities = new GLCapabilities(profile);
		canvas = new GLCanvas(capabilities);
		add(canvas);
		canvas.addGLEventListener(this);
		animator = new FPSAnimator(canvas, framesPerSecond);
		animator.add(canvas);
		setVisible(true);
		addKeyListener(this);
		addWindowListener(this);
	}
	
	public Display() {
		this(600,600);
	}
	
	@Override
	public void init(GLAutoDrawable drawable) {
		GL2 gl = drawable.getGL().getGL2();
		gl.setSwapInterval(1);
		glu = new GLU();
		glut = new GLUT();
		
		gl.glEnable(GL2.GL_DEPTH_TEST);
		gl.glDepthFunc(GL2.GL_LEQUAL);
		gl.glShadeModel(GL2.GL_SMOOTH);
//		gl.glFrontFace(GL2.GL_CULL_FACE);
		gl.glHint(GL2.GL_PERSPECTIVE_CORRECTION_HINT, GL2.GL_NICEST);
		gl.glClearColor(0f, 0f, 0f, 1f);
		
		animator.start();
	}

	@Override
	public void dispose(GLAutoDrawable drawable) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void display(GLAutoDrawable drawable) {
		update(drawable);
		setCamera(drawable);
		render(drawable);
	}
	
	/** Update logical game stuff here */
	private void update(GLAutoDrawable drawable){
		currentFrame++;
		// Each tick causes the block to move down and adds one point as a "staying alive" score
		if (currentFrame >= framesPerSecond){
			if (!gm.moveActiveDown()) {
				gm.settleActive();
			}
			currentFrame = 0;
			++score;
		}
		//let cam look at active piece
//		camLookAt[2] = (height - (0 + gm.getActiveY())) * (2 * Math.PI * radiusOuter) / width - (2 * Math.PI * radiusOuter) / width * currentFrame / framesPerSecond;
		theta += camRotSpeed;
		
		//crude hack
//		if (gm.getActiveY() + gm.getActiveBlock().shape[0].length - 1 >= height){
//			gm.setActiveBlock(gm.randomBlock());
//			System.out.printf("(%s X %s)\n", gm.getActiveBlock().shape.length, gm.getActiveBlock().shape[0].length);
//		}
	}
	
	/** The things to display go here */
	private void render(GLAutoDrawable drawable) {
		GL2 gl = drawable.getGL().getGL2();

		gl.glClear(GL2.GL_COLOR_BUFFER_BIT | GL2.GL_DEPTH_BUFFER_BIT);
		
		gl.glRasterPos2i(50,75);
		glut.glutBitmapString(GLUT.BITMAP_HELVETICA_18, "Score: " + score);

		//draw grid
		gl.glLineWidth(1.1f);
		//renderAsWireframe(drawable, grid);

		//draw blocks
		Color[][] field = gm.getField();
		for (int i = 0; i < field.length; i++) {
			for (int j = 0; j < field[i].length; j++) {
				if (!(field[i][j].equals(Color.black))) {
					Color c = field[i][j];
					double n = 1.0 - (Math.abs(field[i].length/2.0 - j) / (field[i].length/2.0))*0.5;
					gl.glColor3d((c.getRed()/255.0)*n, (c.getGreen()/255.0)*n, (c.getBlue()/255.0)*n);
					gl.glMatrixMode(GL2.GL_MODELVIEW);
					gl.glLoadIdentity();
					gl.glTranslated(0, 0, (height - (i + 1)) * (2 * Math.PI * radiusOuter) / width);
					gl.glRotated(j * 360 / width, 0, 0, 1);
					renderAsTris(drawable, block);
					gl.glColor3d(1, 1, 1);
					gl.glLineWidth(2);
					renderAsWireframe(drawable, block);
				}
			}
		}

		// draw bottom
		for (int j = 0; j < field[0].length; j++) {
			int i = field.length;
			Color c = Color.white;
			double n = 1.0 - (Math.abs(field[0].length/2.0 - j) / (field[0].length/2.0))*0.5;
			gl.glColor3d((c.getRed()/255.0)*n, (c.getGreen()/255.0)*n, (c.getBlue()/255.0)*n);
			gl.glMatrixMode(GL2.GL_MODELVIEW);
			gl.glLoadIdentity();
			gl.glTranslated(0, 0, (height - (i + 1)) * (2 * Math.PI * radiusOuter) / width);
			gl.glRotated(j * 360 / width, 0, 0, 1);
			renderAsTris(drawable, block);
			gl.glColor3d(1, 1, 1);
			gl.glLineWidth(2);
			renderAsWireframe(drawable, block);
		}

		GameBlock activeBlock = gm.getActiveBlock();
		int ax = gm.getActiveX();
		int ay = gm.getActiveY();
		//draw active block
		for (int i = 0; i < activeBlock.shape.length; i++) {
			for (int j = 0; j < activeBlock.shape[i].length; j++) {
				if (activeBlock.shape[i][j]) {
					gl.glMatrixMode(GL2.GL_MODELVIEW);
					gl.glLoadIdentity();
					gl.glTranslated(0, 0, (height - (i + ay)) * (2 * Math.PI * radiusOuter) / width - (2 * Math.PI * radiusOuter) / width * currentFrame / framesPerSecond);
					gl.glRotated((ax + j) * (360 / width), 0, 0, 1);
					gl.glColor3d(activeBlock.color.getRed()/255.0, activeBlock.color.getGreen()/255.0, activeBlock.color.getBlue()/255.0);
					renderAsTris(drawable, block);
					gl.glColor3d(1, 1, 1);
					gl.glLineWidth(2);
					renderAsWireframe(drawable, block);
				}
			}
		}
		//draw models from ArrayList models
		for (int m = 0; m < models.size(); m++){//loop trough world models
			if (models.get(m).triangles.size() > 0){
//				renderAsWireFrame(drawable, models.get(m));
				renderAsTris(drawable, models.get(m));
//				renderAsPoints(drawable, models.get(m));
			}
			else{
				renderAsPoints(drawable, models.get(m));
			}
		}
	}
	
	/** Render a model as triangles */
	private void renderAsTris(GLAutoDrawable drawable, Model model) {
		GL2 gl = drawable.getGL().getGL2();
//		gl.glBegin(GL2.GL_TRIANGLES);
		for (int f = 0; f < model.triangles.size(); f++){//loop though model's triangles (faces)
			for (int j = 0; j < model.triangles.get(f).vertices.length; j++){
				if (model.triangles.get(f).vertices.length == 3){
					gl.glBegin(GL2.GL_TRIANGLES);
//					gl.glColor3d(1, 1.0 / j, 0);
					gl.glVertex3d(
							model.vertices.get(model.triangles.get(f).vertices[j]).getXCoord(), 
							model.vertices.get(model.triangles.get(f).vertices[j]).getYCoord(),
							model.vertices.get(model.triangles.get(f).vertices[j]).getZCoord());
				}
			}
			gl.glEnd();
//			gl.glVertex3d(
//					model.vertices.get(model.triangles.get(f).vertices[0]).getXCoord(), 
//					model.vertices.get(model.triangles.get(f).vertices[0]).getYCoord(),
//					model.vertices.get(model.triangles.get(f).vertices[0]).getZCoord());
//			gl.glColor3d(0, 1, 0);
//			gl.glVertex3d(
//					model.vertices.get(model.triangles.get(f).vertices[1]).getXCoord(), 
//					model.vertices.get(model.triangles.get(f).vertices[1]).getYCoord(), 
//					model.vertices.get(model.triangles.get(f).vertices[1]).getZCoord());
//			gl.glColor3d(1, 1, 0);
//			gl.glVertex3d(
//					model.vertices.get(model.triangles.get(f).vertices[2]).getXCoord(), 
//					model.vertices.get(model.triangles.get(f).vertices[2]).getYCoord(), 
//					model.vertices.get(model.triangles.get(f).vertices[2]).getZCoord());
		}
//		gl.glEnd();
	}
	
	/** Render a model as vertices */
	private void renderAsPoints(GLAutoDrawable drawable, Model model) {
		GL2 gl = drawable.getGL().getGL2();
		gl.glPointSize(2);
		gl.glBegin(GL2.GL_POINTS);
		for (int v = 0; v < model.vertices.size(); v++){//loop though model's vertices (points)
			gl.glColor3d(1, 1, 1);
			gl.glVertex3d(
					model.vertices.get(v).getXCoord(),
					model.vertices.get(v).getYCoord(),
					model.vertices.get(v).getZCoord());
		}
		gl.glEnd();
	}

	/** Render a model as wireframe */
	private void renderAsWireframe(GLAutoDrawable drawable, Model model) {
		GL2 gl = drawable.getGL().getGL2();
		gl.glBegin(GL2.GL_LINES);
		for (int f = 0; f < model.triangles.size(); f++){//loop thought model's triangles
			gl.glColor3d(.5, .5, .5);
//			gl.glBegin(GL2.GL_LINE_LOOP); //why does this not work?
			for (int v = 0; v < model.triangles.get(f).vertices.length; v++){
				gl.glVertex3d(
						model.vertices.get(model.triangles.get(f).vertices[v]).getXCoord(),
						model.vertices.get(model.triangles.get(f).vertices[v]).getYCoord(),
						model.vertices.get(model.triangles.get(f).vertices[v]).getZCoord());
			}
//			gl.glEnd();
		}
		gl.glEnd();
	}
	
	private void setCamera(GLAutoDrawable drawable) {
		GL2 gl = drawable.getGL().getGL2();
		gl.glMatrixMode(GL2.GL_PROJECTION);
		gl.glLoadIdentity();
		
		glu.gluPerspective(45, 1, 1, 1000);
		glu.gluLookAt(
				camDistFromOrigo * Math.cos(theta),//cam's x-position 
				camDistFromOrigo * Math.sin(theta),//cam's y-position
				camHeight, //cam's z-position
				camLookAt[0], //cam target x
				camLookAt[1], //cam target y
				camLookAt[2], //cam target z
				0, 0, 1);//what is up?
		
		gl.glMatrixMode(GL2.GL_MODELVIEW);
		gl.glLoadIdentity();
	}
	
	@Override
	public void reshape(GLAutoDrawable drawable, int x, int y, int width,
			int height) {
		GL2 gl = drawable.getGL().getGL2();
		gl.glViewport(x, y, width, height);
	}
	
	/*
	private Model buildGrid(){
		Model model = new Model();
		double xPos = 0;
		double yPos = 0;
		double zPos = 0;
		int numOfBlocks = width;
		int numOfLevels = height + 1;//antal etager
		for (int j = 0; j < numOfLevels; j++){
			for (int i = 0; i < numOfBlocks; i++){
				xPos = radiusOuter * Math.cos(2 * Math.PI * i / numOfBlocks);
				yPos = radiusOuter * Math.sin(2 * Math.PI * i / numOfBlocks);
				zPos = j * (2 * Math.PI * radiusOuter) / numOfBlocks;
				model.addVertex(new Vertex(xPos, yPos, zPos));//add outer vertex
				xPos = radiusInner * Math.cos(2 * Math.PI * i / numOfBlocks);
				yPos = radiusInner * Math.sin(2 * Math.PI * i / numOfBlocks);
				zPos = j * (2 * Math.PI * radiusOuter) / numOfBlocks;
				model.addVertex(new Vertex(xPos, yPos, zPos));//add inner vertex
				model.addTriangle(new Triangle(
						(j * 2 * numOfBlocks) + ((i * 2 + 2 * numOfBlocks) % (2 * numOfBlocks)), (j * 2 * numOfBlocks) + (((i * 2 + 2 * numOfBlocks + 2)) % (2 * numOfBlocks))));//outer circles
//				model.addTriangle(new Triangle(
//						(j * 2 * numOfBlocks) + ((i * 2 + 2 * numOfBlocks) % (2 * numOfBlocks)), (j * 2 * numOfBlocks) + (((i * 2 + 2 * numOfBlocks + 1)) % (2 * numOfBlocks))));//rays
//				model.addTriangle(new Triangle(
//						(j * 2 * numOfBlocks) + ((i * 2 + 2 * numOfBlocks + 1) % (2 * numOfBlocks)), (j * 2 * numOfBlocks) + (((i * 2 + 2 * numOfBlocks + 3)) % (2 * numOfBlocks))));//inner circles
				if (j == numOfLevels - 1){
					model.addTriangle(new Triangle(
							(j * 2 * numOfBlocks) + ((i * 2)), (i * 2)));// + 2 * numOfBlocks) % (2 * numOfBlocks)));//outer vertical lines
//					model.addTriangle(new Triangle(
//							(j * 2 * numOfBlocks) + (i * 2 + 1), i * 2 + 1));//inner vertical lines
				}
			}
			
		}
		return model;
	}
	//*/
	
	private Model buildBlock(){
		Model model = new Model();
//		model.setLocation(radiusOuter * Math.cos(2 * Math.PI * 0 / width), 0, 0);
		
		model.addVertex(new Vertex(radiusOuter * Math.cos(2 * Math.PI * 0 / width), radiusOuter * Math.sin(2 * Math.PI * 0 / width), 0 * (2 * Math.PI * radiusOuter) / width));//v0
		model.addVertex(new Vertex(radiusOuter * Math.cos(2 * Math.PI * 1 / width), radiusOuter * Math.sin(2 * Math.PI * 1 / width), 0 * (2 * Math.PI * radiusOuter) / width));//v1
		model.addVertex(new Vertex(radiusInner * Math.cos(2 * Math.PI * 0 / width), radiusInner * Math.sin(2 * Math.PI * 0 / width), 0 * (2 * Math.PI * radiusOuter) / width));//v2
		model.addVertex(new Vertex(radiusInner * Math.cos(2 * Math.PI * 1 / width), radiusInner * Math.sin(2 * Math.PI * 1 / width), 0 * (2 * Math.PI * radiusOuter) / width));//v3

		model.addVertex(new Vertex(radiusOuter * Math.cos(2 * Math.PI * 0 / width), radiusOuter * Math.sin(2 * Math.PI * 0 / width), 1 * (2 * Math.PI * radiusOuter) / width));//v4
		model.addVertex(new Vertex(radiusOuter * Math.cos(2 * Math.PI * 1 / width), radiusOuter * Math.sin(2 * Math.PI * 1 / width), 1 * (2 * Math.PI * radiusOuter) / width));//v5
		model.addVertex(new Vertex(radiusInner * Math.cos(2 * Math.PI * 0 / width), radiusInner * Math.sin(2 * Math.PI * 0 / width), 1 * (2 * Math.PI * radiusOuter) / width));//v6
		model.addVertex(new Vertex(radiusInner * Math.cos(2 * Math.PI * 1 / width), radiusInner * Math.sin(2 * Math.PI * 1 / width), 1 * (2 * Math.PI * radiusOuter) / width));//v7
		
		//bottom
		model.addTriangle(new Triangle(0, 1, 2));
		model.addTriangle(new Triangle(1, 2, 3));
		//top
		model.addTriangle(new Triangle(4, 5, 6));
		model.addTriangle(new Triangle(5, 6, 7));
		//front
		model.addTriangle(new Triangle(0, 1, 4));
		model.addTriangle(new Triangle(1, 4, 5));
		//back
		model.addTriangle(new Triangle(2, 3, 6));
		model.addTriangle(new Triangle(3, 6, 7));
		//side
		model.addTriangle(new Triangle(0, 4, 2));
		model.addTriangle(new Triangle(4, 2, 6));
		//other side
		model.addTriangle(new Triangle(1, 5, 7));
		model.addTriangle(new Triangle(1, 3, 7));
		return model;
	}

	@Override
	public void blockUsedEvent() {
		gm.setActiveBlock(gm.randomBlock());
		// Placing a block yields a few points and resets bonus to nothing
		score += 5;
		bonus = 0;
	}

	@Override
	public void removeRowEvent(int row) {
		// ToDo: Show some fancy effects blowing up the row
		// A row clear event yields many points and increases the bonus so if you clear 5 rows at once you get
		// width*10 + width*11 + width*12 + width*13 + width*14 points
		score += width*(10+bonus);
		++bonus;
	}

	@Override
	public void gameOverEvent() {
		System.out.println("Final score: " + score);
		System.exit(0);
	}

	public void keyPressed(KeyEvent e) {
		if (e.getKeyCode() == KeyEvent.VK_UP
				|| e.getKeyCode() == KeyEvent.VK_KP_UP
				|| e.getKeyCode() == KeyEvent.VK_W) {
			gm.rotateActiveLeft();
		} else if (e.getKeyCode() == KeyEvent.VK_DOWN
				|| e.getKeyCode() == KeyEvent.VK_KP_DOWN
				|| e.getKeyCode() == KeyEvent.VK_S) {
			gm.moveActiveDown();
			currentFrame = 0;
		} else if (e.getKeyCode() == KeyEvent.VK_LEFT
				|| e.getKeyCode() == KeyEvent.VK_KP_LEFT
				|| e.getKeyCode() == KeyEvent.VK_A) {
			gm.moveActiveLeft();
		} else if (e.getKeyCode() == KeyEvent.VK_RIGHT
				|| e.getKeyCode() == KeyEvent.VK_KP_RIGHT
				|| e.getKeyCode() == KeyEvent.VK_D) {
			gm.moveActiveRight();
		}
	}

	public void keyReleased(KeyEvent e) {
	}

	public void keyTyped(KeyEvent e) {
		if (e.getKeyChar() == 'w' || e.getKeyChar() == 'W') {
			gm.rotateActiveLeft();
		} else if (e.getKeyChar() == 's' || e.getKeyChar() == 'S') {
			gm.moveActiveDown();
			currentFrame = 0;
		} else if (e.getKeyChar() == 'a' || e.getKeyChar() == 'A') {
			gm.moveActiveLeft();
		} else if (e.getKeyChar() == 'd' || e.getKeyChar() == 'D') {
			gm.moveActiveRight();
		} else if (e.getKeyChar() == ' ') {
			gm.settleActive();
			currentFrame = 0;
		}
	}

	@Override
	public void windowActivated(WindowEvent e) {
		System.out.println("hello");
	}

	@Override
	public void windowClosed(WindowEvent e) {
		exit(0);
	}

	@Override
	public void windowClosing(WindowEvent e) {
		exit(0);
	}

	@Override
	public void windowDeactivated(WindowEvent e) {
	}

	@Override
	public void windowDeiconified(WindowEvent e) {
	}

	@Override
	public void windowIconified(WindowEvent e) {
	}

	@Override
	public void windowOpened(WindowEvent e) {
	}
	
	private void exit(int status){
		animator.stop();
		canvas.destroy();
		System.out.println("Goodbye");
		System.exit(status);
	}

	public static void main(String[] args){
		new Display(600, 600);
	}
}
